# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

from .trainer import DistributedTrainer, set_random_seed
from .args import get_args
# from .dist_launcher import initialize_distributed,kill_children
# from ._utils import batch_to,batch_apply
