# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

import os
# This statement must be executed at the very beginning, i.e. before import torch
os.environ["OMP_NUM_THREADS"] = "1"
