# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# zbo@zju.edu.cn
# 2022-08-08
# ============================================================================

class BertTokenizer(object):
    def __init__(self, **kwargs,) -> None:
        pass

class Vocab(object):
    def __init__(self, **kwargs,) -> None:
        pass

    @staticmethod
    def from_dict(vocab):
        return None